/*
This is a part of the IconDesk source code.

Copyright (C) 2003 Erik Christiansson, aka Sci
www.alphafish.com
erik@alphafish.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "Group.h"

Group::Group(const char* nameIn, HWND parentWnd)
{
	// Store input
	name = (char*)malloc(strlen(nameIn) + 1);
	memcpy(name, nameIn, strlen(nameIn) + 1);
	_parentWnd = parentWnd;

	// Bind the icon extraction function
	PrivateExtractIcons = (UINT (WINAPI *)(LPCTSTR,int,int,int,HICON*,UINT*,UINT,UINT)) GetProcAddress(GetModuleHandle("USER32.DLL"), "PrivateExtractIconsA");
	if(!PrivateExtractIcons) {
		DWORD dw = GetLastError();
		LogPrintf(LOG_ERROR, "Could not bind PrivateExtractIcons (Error: %u)", dw);
	}

	// We have not loaded our icons yet
	_iconsLoaded = FALSE;

	// A general buffer used to read the settings
	char buffer[MAX_LINE_LENGTH];
	// Handle to step.rc
	FILE* fpStep;

	// Get group metrics
	_visible = _getBool("Visible", TRUE);
	_iconX = _getInt("X", 0);
	_iconY = _getInt("Y", 0);
	if(_iconX < 0) _iconX += GetSystemMetrics(SM_CXSCREEN);
	if(_iconY < 0) _iconY += GetSystemMetrics(SM_CYSCREEN);
	_iconSize = _getInt("Size", 32);
	_iconSpacingX = _getInt("SpacingX", 0);
	_iconSpacingY = _getInt("SpacingY", 0);

	_iconHideExtension = _getBool("HideExtension", TRUE);
	_iconFoldersFirst = _getBool("FoldersFirst", TRUE);
	_iconShowHidden = _getBool("ShowHidden", FALSE);
	_showTooltips = _getBool("Tooltips", TRUE);

	// Get text settings
	_hideText = _getBool("HideText", FALSE);
	if(!_hideText) {
		// Get the background brush
		_getString("TextBackground", buffer, "", MAX_LINE_LENGTH);
		if(strlen(buffer)) {
			_textBackground = CreateSolidBrush(_getColor("TextBackground", RGB(255,255,255)));
		}
		else {
			_textBackground = NULL;
		}

		// Text box metrics
		_textSpacing = _getInt("TextSpacing", 5);
		_textHeight = _getInt("TextHeight", 28);
		_textWidth = _getInt("TextWidth", 64);
	
		// Relative position of the text box
		_getString("TextPosition", buffer, "", MAX_LINE_LENGTH);
		if(!stricmp(buffer, "right")) {
			_textPosition = DD_RIGHT;
		}
		else if(!stricmp(buffer, "left")) {
			_textPosition = DD_LEFT;
		}
		else if(!stricmp(buffer, "top")) {
			_textPosition = DD_UP;
		}
		else {
			_textPosition = DD_DOWN;
		}
	
		// Text alignment
		_getString("TextAlign", buffer, "", MAX_LINE_LENGTH);
		if(!stricmp(buffer, "right")) {
			_textAlign = DT_RIGHT;
		}
		else if(!stricmp(buffer, "left")) {
			_textAlign = DT_LEFT;
		}
		else {
			_textAlign = DT_CENTER;
		}

		// Get the font
		_textColor = _getColor("FontColor", RGB(255, 255, 255));
		_getString("Font", buffer, "Arial", MAX_LINE_LENGTH);
		int fontSize = _getInt("FontSize", 14);
		BOOL fontItalic = _getBool("FontItalic", FALSE);
		BOOL fontBold = _getBool("FontBold", FALSE);
		_textFont = CreateFont(fontSize, 0, 0, 0, fontBold ? FW_BOLD : FW_NORMAL, fontItalic, FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, buffer);
		if(!_textFont) {
			LogPrintf(LOG_ERROR, "Could not load font \"%s\"", buffer);
		}

		// Text shadow settings
		_textShadowColor = _getColor("TextShadowColor", RGB(80, 80, 80));
		_textShadowDepth = _getInt("TextShadowDepth", 0);

		// Text outline settings
		_textOutlineColor = _getColor("TextOutlineColor", RGB(40, 40, 40));
		_textOutline = _getBool("TextOutline", FALSE);
		if(_textOutline) {
			if(_textShadowDepth > 0) {
				_textShadowDepth++;
			}
			else if(_textShadowDepth < 0) {
				_textShadowDepth--;
			}
		}
	}
	else {
		_textWidth = 0;
		_textHeight = 0;
		_textSpacing = 0;
	}

	// Icon direction
	_getString("Direction", buffer, "", MAX_LINE_LENGTH);
	if(!stricmp(buffer, "down")) {
		_iconDirection = DD_DOWN;
	}
	else if(!stricmp(buffer, "left")) {
		_iconDirection = DD_LEFT;
	}
	else if(!stricmp(buffer, "up")) {
		_iconDirection = DD_UP;
	}
	else {
		_iconDirection = DD_RIGHT;
	}

	// Icon wraping
	_getString("WrapDirection", buffer, "", MAX_LINE_LENGTH);
	if(_iconDirection == DD_DOWN || _iconDirection == DD_UP) {
		if(!stricmp(buffer, "left")) {
			_iconWrapDirection = DD_LEFT;
		}
		else {
			_iconWrapDirection = DD_RIGHT;
		}
	}
	else {
		if(!stricmp(buffer, "up")) {
			_iconWrapDirection = DD_UP;
		}
		else {
			_iconWrapDirection = DD_DOWN;
		}
	}
	_iconWrapCount = max(_getInt("WrapCount", 10), 1);

	// Set the default mouse actions
	UINT nMask;
	nMask = ME_LBUTTONDBL;
	_mouseEvents.insert(MouseMap::value_type(nMask, ".open"));
	nMask |= ME_MENU;
	_mouseEvents.insert(MouseMap::value_type(nMask, ".properties"));
	nMask = ME_RBUTTONUP;
	_mouseEvents.insert(MouseMap::value_type(nMask, ".context"));

	// Read mouse actions
	fpStep = LCOpen(NULL);
	if(fpStep) {
		char    szPrefix[1024];
		char	szBuffer[4096];
		char	szDummy[4096], szEvent[4096], szModifiers[4096], szCommand[4096];
		char*	tokens[3];

		tokens[0] = szDummy;
		tokens[1] = szEvent;
		tokens[2] = szModifiers;

		buffer[0] = 0;

		sprintf(szPrefix, "*IconDesk%sMouse", name);

		while(LCReadNextConfig(fpStep, szPrefix, szBuffer, sizeof(szBuffer))) {
			int count;

			szDummy[0] = szEvent[0] = szModifiers[0] = szCommand[0] = '\0';

			count = LCTokenize(szBuffer, tokens, 3, szCommand);

			if(count == 3) {
				UINT nMask = 0;

				// Parse mouse events
				if(!stricmp(szEvent, "LeftUp")) {
					nMask |= ME_LBUTTONUP;
				}
				else if(!stricmp(szEvent, "LeftDown")) {
					nMask |= ME_LBUTTONDOWN;
				}
				else if(!stricmp(szEvent, "LeftDbl")) {
					nMask |= ME_LBUTTONDBL;
				}
				else if(!stricmp(szEvent, "MiddleUp")) {
					nMask |= ME_MBUTTONUP;
				}
				else if(!stricmp(szEvent, "MiddleDown")) {
					nMask |= ME_MBUTTONDOWN;
				}
				else if(!stricmp(szEvent, "MiddleDbl")) {
					nMask |= ME_MBUTTONDBL;
				}
				else if(!stricmp(szEvent, "RightUp")) {
					nMask |= ME_RBUTTONUP;
				}
				else if(!stricmp(szEvent, "RightDown")) {
					nMask |= ME_RBUTTONDOWN;
				}
				else if(!stricmp(szEvent, "RightDbl")) {
					nMask |= ME_RBUTTONDBL;
				}
				else {
					// Notify the user that an invalid mouse event was found
					char szMessage[1024];
					sprintf(szMessage, "\"%s\" is not a valid mouse even", szEvent);
					MessageBox(NULL, szMessage, "IconDesk", MB_SETFOREGROUND);
				}

				// Parse modifier keys
				char* szKey = strtok(szModifiers, "+");
				while(szKey) {
					if(!stricmp(szKey, "CTRL")) {
						nMask |= ME_CONTROL;
					}
					else if(!stricmp(szKey, "ALT")) {
						nMask |= ME_MENU;
					}
					else if(!stricmp(szKey, "SHIFT")) {
						nMask |= ME_SHIFT;
					}
					else if(!stricmp(szKey, "APPS")) {
						nMask |= ME_APPS;
					}
					else if(stricmp(szKey, ".none")) {
						// Notify the user that an invalid modifier was found
						char szMessage[1024];
						sprintf(szMessage, "\"%s\" is not a valid modifier key", szKey);
						MessageBox(NULL, szMessage, "IconDesk", MB_SETFOREGROUND);
					}
					szKey = strtok(NULL, "+");
				}

				MouseMapIter it;
				it = _mouseEvents.find(nMask);
				if(it != _mouseEvents.end()) {
					_mouseEvents.erase(it);
				}
				_mouseEvents.insert(MouseMap::value_type(nMask, szCommand));
			}
		}
	}
	LCClose(fpStep);

	// Icon effects
	_effectFlags = 0;
	_hueIntensity = (UCHAR)min(max(_getInt("HueIntensity", 0), 0), 255);
	_hueColor = _getColor("HueColor", RGB(128, 128, 128));
	if(_hueIntensity) {
		_effectFlags |= FX_HUE;
	}
	_saturation = (UCHAR)min(max(_getInt("Saturation", 255), 0), 255);
	if(_saturation != 255) {
		_effectFlags |= FX_SATURATION;
	}

	// Hide patterns
	fpStep = LCOpen(NULL);
	if(fpStep) {
		char szPrefix[MAX_LINE_LENGTH], szBuffer[MAX_LINE_LENGTH], szData[MAX_LINE_LENGTH], szDummy[MAX_LINE_LENGTH];
		char* rszTokens[1];

		rszTokens[0] = szDummy;
		sprintf(szPrefix, "*IconDesk%sHide", name);

		szBuffer[0] = 0;
		while(LCReadNextConfig(fpStep, szPrefix, szBuffer, sizeof(szBuffer))) {
			szDummy[0] = szData[0] = 0;
			LCTokenize(szBuffer, rszTokens, 1, szData);

			_patternHide.push_back(szData);
		}
		LCClose(fpStep);
	}

	// First patterns
	fpStep = LCOpen(NULL);
	if(fpStep) {
		char szPrefix[MAX_LINE_LENGTH], szBuffer[MAX_LINE_LENGTH], szData[MAX_LINE_LENGTH], szDummy[MAX_LINE_LENGTH];
		char* rszTokens[1];

		rszTokens[0] = szDummy;
		sprintf(szPrefix, "*IconDesk%sFirst", name);

		szBuffer[0] = 0;
		while(LCReadNextConfig(fpStep, szPrefix, szBuffer, sizeof(szBuffer))) {
			szDummy[0] = szData[0] = 0;
			LCTokenize(szBuffer, rszTokens, 1, szData);

			_patternFirst.push_back(szData);
		}
		LCClose(fpStep);
	}

	// Last patterns
	fpStep = LCOpen(NULL);
	if(fpStep) {
		char szPrefix[MAX_LINE_LENGTH], szBuffer[MAX_LINE_LENGTH], szData[MAX_LINE_LENGTH], szDummy[MAX_LINE_LENGTH];
		char* rszTokens[1];

		rszTokens[0] = szDummy;
		sprintf(szPrefix, "*IconDesk%sLast", name);

		szBuffer[0] = 0;
		while(LCReadNextConfig(fpStep, szPrefix, szBuffer, sizeof(szBuffer))) {
			szDummy[0] = szData[0] = 0;
			LCTokenize(szBuffer, rszTokens, 1, szData);

			_patternLast.push_back(szData);
		}
		LCClose(fpStep);
	}

	// Load the folders
	fpStep = LCOpen(NULL);
	if(fpStep) {
		char szPrefix[MAX_LINE_LENGTH], szBuffer[MAX_LINE_LENGTH], szData[MAX_LINE_LENGTH], szDummy[MAX_LINE_LENGTH];
		char* rszTokens[1];

		rszTokens[0] = szDummy;
		sprintf(szPrefix, "*IconDesk%sFolder", name);

		szBuffer[0] = 0;
		while(LCReadNextConfig(fpStep, szPrefix, szBuffer, sizeof(szBuffer))) {
			szDummy[0] = szData[0] = 0;
			LCTokenize(szBuffer, rszTokens, 1, szData);

			// Remove quotes around the path
			if(szData[0] == '"' && szData[strlen(szData) - 1] == '"') {
				strcpy(szData, szData + 1);
				szData[strlen(szData) - 1] = '\0';
			}

			_folders.push_back(szData);

			LogPrintf(LOG_DEBUG, "Folder \"%s\" \"%s\"", name, szData);

			// Folder watcher
			LPITEMIDLIST pidl = NULL;
			IShellFolder *psfDesktop = NULL;
			HRESULT hr;
			WCHAR wszFolder[MAX_PATH];

			if(!stricmp(szData, ".desktop")) {
				LPMALLOC pMalloc;
				if(SUCCEEDED(SHGetMalloc(&pMalloc))) {
					void *temp = pMalloc->Alloc(2);
				    pMalloc->Release();
					ZeroMemory(temp, 2);
					pidl = (LPITEMIDLIST)temp;
				}
			}
			else {
				// Get the IShellFolder for the folder
				MultiByteToWideChar(CP_ACP, 0, szData, -1, wszFolder, sizeof(wszFolder) / sizeof(WCHAR));
				// Used to bind the interface
				hr = SHGetDesktopFolder(&psfDesktop);
				// Get the PIDL
				hr = psfDesktop->ParseDisplayName(NULL, NULL, wszFolder, NULL, &pidl, NULL);
				// Release the desktop
				psfDesktop->Release();
			}
			_notifiers.push_back(new ShellNotify(pidl, this));
		}
		LCClose(fpStep);
	}

	// If no folder was specified, use the desktop
	if(_folders.size() == 0) {
		// The default folder is the desktop
		_folders.push_back(".desktop");

		// Create a folder watcher
		LPITEMIDLIST pidl;
		LPMALLOC pMalloc;
		if(SUCCEEDED(SHGetMalloc(&pMalloc))) {
			void *temp = pMalloc->Alloc(2);
			pMalloc->Release();
			ZeroMemory(temp, 2);
			pidl = (LPITEMIDLIST)temp;

			// Folder watcher
			_notifiers.push_back(new ShellNotify(pidl, this));
		}
	}
}

Group::~Group()
{
	_deleteIcons();

	if(_textFont) {
		DeleteObject(_textFont);
	}
	if(_textBackground) {
		DeleteObject(_textBackground);
	}

	for(ShellNotifyListIter it = _notifiers.begin(); it != _notifiers.end(); it++) {
		delete (*it);
	}

	free(name);
}

void Group::show()
{
	_loadTooltips();
	_visible = TRUE;
	InvalidateRect(_parentWnd, NULL, TRUE);
}

void Group::hide()
{
	_deleteTooltips();
	_visible = FALSE;
	InvalidateRect(_parentWnd, NULL, TRUE);
}

void Group::toggle()
{
	if(_visible) {
		hide();
	}
	else {
		show();
	}
}

void Group::update()
{
	_deleteIcons();
	_loadIcons();
	InvalidateRect(_parentWnd, NULL, TRUE);
}

BOOL Group::mouseEvent(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	if(!_visible) {
		return FALSE;
	}

	Icon* clicked = _getIconAtPt(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));

	if(clicked) {
		if(uMsg == WM_RBUTTONDOWN || uMsg == WM_LBUTTONDOWN || uMsg == WM_MBUTTONDOWN) {
			if(_selectedIcon) {
				InvalidateRect(_parentWnd, &_selectedIcon->textRect, TRUE);
			}
			_selectedIcon = clicked;
			if(_selectedIcon) {
				InvalidateRect(_parentWnd, &_selectedIcon->textRect, TRUE);
			}
		}

		UINT nMask = 0;

		if(GetAsyncKeyState(VK_SHIFT) & 0x8000) {
			nMask |= ME_SHIFT;
		}

		if(GetAsyncKeyState(VK_CONTROL) & 0x8000) {
			nMask |= ME_CONTROL;
		}

		if(GetAsyncKeyState(VK_MENU) & 0x8000) {
			nMask |= ME_MENU;
		}

		if(GetAsyncKeyState(VK_APPS) & 0x8000) {
			nMask |= ME_APPS;
		}

		switch(uMsg) {
			case WM_LBUTTONUP:
				nMask |= ME_LBUTTONUP;
				break;

			case WM_LBUTTONDOWN:
				nMask |= ME_LBUTTONDOWN;
				break;

			case WM_LBUTTONDBLCLK:
				nMask |= ME_LBUTTONDBL;
				break;

			case WM_MBUTTONUP:
				nMask |= ME_MBUTTONUP;
				break;

			case WM_MBUTTONDOWN:
				nMask |= ME_MBUTTONDOWN;
				break;

			case WM_MBUTTONDBLCLK:
				nMask |= ME_MBUTTONDBL;
				break;

			case WM_RBUTTONUP:
				nMask |= ME_RBUTTONUP;
				break;

			case WM_RBUTTONDOWN:
				nMask |= ME_RBUTTONDOWN;
				break;

			case WM_RBUTTONDBLCLK:
				nMask |= ME_RBUTTONDBL;
				break;

		}

		MouseMapIter it;
		it = _mouseEvents.find(nMask);
		if(it != _mouseEvents.end()) {
			clicked->Execute((*it).second.c_str());
        }
		return TRUE;
	}
	else {
		if(_selectedIcon) {
			InvalidateRect(_parentWnd, &_selectedIcon->textRect, TRUE);
		}
		_selectedIcon = NULL;
		return FALSE;
	}
}

void Group::_deleteIcons()
{
	for(IconListIter it = _iconList.begin(); it != _iconList.end(); it++) {
		if((*it)->toolId) {
			DeleteTooltip((*it)->toolId);
		}

		delete (*it);
	}
	_iconList.clear();
	_iconsLoaded = FALSE;
}

void Group::_loadTooltips()
{
	if(_showTooltips) {
		for(IconListIter it = _iconList.begin(); it != _iconList.end(); it++) {
			(*it)->toolId = AddTooltip((*it)->title, (*it)->iconRect, (*it)->toolId);
		}
	}
}

void Group::_deleteTooltips()
{
	if(_showTooltips) {
		for(IconListIter it = _iconList.begin(); it != _iconList.end(); it++) {
			if((*it)->toolId) {
				DeleteTooltip((*it)->toolId);
			}
		}
	}
}

void Group::_loadFolder(const char* szFolder)
{
    LPMALLOC pMalloc;
    LPITEMIDLIST pidlFolder = NULL;
    LPITEMIDLIST pidlItems = NULL;
    IShellFolder *psfDesktop = NULL;
    IShellFolder *psfFolder = NULL;
    LPENUMIDLIST ppenum = NULL;
    ULONG celtFetched;
    HRESULT hr;
    STRRET strDispName;
    ULONG uAttr;
	WCHAR wszFolder[MAX_PATH];
	BOOL bFilesystem = TRUE;
	char szMyDocuments[1024];

	// For memory managment
    hr = SHGetMalloc(&pMalloc);

	if(!stricmp(szFolder, ".desktop")) {
	    hr = SHGetDesktopFolder(&psfFolder);
		// Create a pidl with only the terminating NULLs
		pidlFolder = (LPITEMIDLIST)pMalloc->Alloc(2);
		ZeroMemory(pidlFolder, 2);
		bFilesystem = FALSE;
	}
	else {
		// Get the IShellFolder for the folder
		MultiByteToWideChar(CP_ACP, 0, szFolder, -1, wszFolder, sizeof(wszFolder) / sizeof(WCHAR));
		// Used to bind the interface
		hr = SHGetDesktopFolder(&psfDesktop);
		// Get the PIDL
		hr = psfDesktop->ParseDisplayName(NULL, NULL, wszFolder, NULL, &pidlFolder, NULL);
		// Get the interface
		hr = psfDesktop->BindToObject(pidlFolder, NULL, IID_IShellFolder, (LPVOID *)&psfFolder);
		// Is it a filesystem folder?
		uAttr = SFGAO_FILESYSTEM;
        psfDesktop->GetAttributesOf(1, (LPCITEMIDLIST*)&pidlFolder, &uAttr);
		if((uAttr & SFGAO_FILESYSTEM) == NULL) {
			bFilesystem = FALSE;
		}
		// Release the desktop
		psfDesktop->Release();
	}

	// Ugly hack for My Documents
	if(!bFilesystem) {
		DWORD dwDataSize = 1024;
		SHGetValue(HKEY_CURRENT_USER, "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders", "Personal", NULL, szMyDocuments, &dwDataSize);
	}

	// Get the enumeration object
	DWORD enumFlags = SHCONTF_FOLDERS | SHCONTF_NONFOLDERS;
	if(_iconShowHidden) {
		enumFlags |= SHCONTF_INCLUDEHIDDEN;
	}
    hr = psfFolder->EnumObjects(NULL, enumFlags, &ppenum);
    while(ppenum->Next(1, &pidlItems, &celtFetched) == S_OK && (celtFetched) == 1) {
		// Create a icon object
		FileIcon* icon = new FileIcon();

		icon->hParent = _parentWnd;

		// Get the title
		uAttr = SFGAO_FILESYSTEM;
        psfFolder->GetAttributesOf(1, (LPCITEMIDLIST*)&pidlItems, &uAttr);
        psfFolder->GetDisplayNameOf(pidlItems, SHGDN_INFOLDER, &strDispName);
		_StrRetToBuf(&strDispName, pidlItems, icon->title, MAX_PATH);

		// Get the filename
        psfFolder->GetDisplayNameOf(pidlItems, SHGDN_NORMAL | SHGDN_FORPARSING, &strDispName);
        _StrRetToBuf(&strDispName, pidlItems, icon->filename, MAX_PATH);

		// Ugly hack to fix the My Documents icon
		if(!bFilesystem) {
			if(!strcmp(icon->filename, szMyDocuments)) {
				strcpy(icon->filename, "::{450D8FBA-AD25-11D0-98A8-0800361B1103}");
			}
		}

		// Get the full PIDL (relative to the desktop)
		icon->SetPIDL(PIDLAppend(pidlFolder, pidlItems, pMalloc));

		if(_iconHideExtension) {
			PathRemoveExtension(icon->title);
		}

		// Is it a folder?
		uAttr = SFGAO_FOLDER;
        psfFolder->GetAttributesOf(1, (LPCITEMIDLIST*)&pidlItems, &uAttr);
		if(uAttr & SFGAO_FOLDER) {
			icon->lFlags = icon->lFlags | Icon::ICON_FOLDER;
		}

		// Release the PIDL
        pMalloc->Free(pidlItems);

		for(StringListIter first = _patternFirst.begin(); first != _patternFirst.end(); first++) {
			if(match((*first).c_str(), icon->title) || match((*first).c_str(), icon->filename)) {
				icon->lFlags |= Icon::ICON_FIRST;
				break;
			}
		}
		for(StringListIter last = _patternLast.begin(); last != _patternLast.end(); last++) {
			if(match((*last).c_str(), icon->title) || match((*last).c_str(), icon->filename)) {
				icon->lFlags |= Icon::ICON_LAST;
				break;
			}
		}

		// If the icon title matches any of the hide patterns, don't show it
		BOOL bHide = FALSE;
		for(StringListIter hide = _patternHide.begin(); hide != _patternHide.end(); hide++) {
			bHide = match((*hide).c_str(), icon->title) || match((*hide).c_str(), icon->filename);
			if(bHide) {
				break;
			}
		}

		if(!bHide) {
			// Add it to the list
			_iconList.push_back((Icon*)icon);
		}
		else {
			delete icon;
		}
    }

	// Release interfaces/PIDLs
    ppenum->Release();
    pMalloc->Free(pidlFolder);
    pMalloc->Release();
    psfFolder->Release();
}

void Group::_loadIcons()
{
	_iconsLoaded = TRUE;

	for(StringListIter slit = _folders.begin(); slit != _folders.end(); slit++) {
		_loadFolder((*slit).c_str());
	}

	_iconList.sort(Group::_compareIcons);
	if(_iconFoldersFirst) {
		_iconList.sort(Group::_compareIconsFolder);
	}
	_iconList.sort(Group::_compareIconsPrio);

	// Calculate RECT's and load icons
	int i = 0, x = 0, y = 0;
	int boxWidth = 0, boxHeight = 0;
	int offsetIconX = 0, offsetIconY = 0;
	int offsetTextX = 0, offsetTextY = 0;

	if(_hideText) {
		boxWidth = _iconSize;
		boxHeight = _iconSize;
		offsetIconX = 0;
		offsetIconY = 0;
		offsetTextX = 0;
		offsetTextY = 0;
	}
	else {
		switch(_textPosition) {
		case DD_DOWN:
			boxWidth = max(_iconSize, _textWidth);
			boxHeight = _iconSize + _textSpacing + _textHeight;
			offsetIconX = max(_iconSize, _textWidth) / 2 - _iconSize / 2;
			offsetIconY = 0;
			offsetTextX = max(_iconSize, _textWidth) / 2 - _textWidth / 2;
			offsetTextY = _iconSize + _textSpacing;
			break;

		case DD_UP:
			boxWidth = max(_iconSize, _textWidth);
			boxHeight = _iconSize + _textSpacing + _textHeight;
			offsetIconX = max(_iconSize, _textWidth) / 2 - _iconSize / 2;
			offsetIconY = _textHeight + _textSpacing;
			offsetTextX = max(_iconSize, _textWidth) / 2 - _textWidth / 2;
			offsetTextY = 0;
			break;

		case DD_RIGHT:
			boxWidth = _iconSize + _textSpacing + _textWidth;
			boxHeight = max(_iconSize, _textHeight);
			offsetIconX = 0;
			offsetIconY = max(_iconSize, _textHeight) / 2 - _iconSize / 2;
			offsetTextX = _iconSize + _textSpacing;
			offsetTextY = max(_iconSize, _textHeight) / 2 - _textHeight / 2;
			break;

		case DD_LEFT:
			boxWidth = _iconSize + _textSpacing + _textWidth;
			boxHeight = max(_iconSize, _textHeight);
			offsetIconX = _textWidth + _textSpacing;
			offsetIconY = max(_iconSize, _textHeight) / 2 - _iconSize / 2;
			offsetTextX = 0;
			offsetTextY = max(_iconSize, _textHeight) / 2 - _textHeight / 2;
			break;

		}
	}

	IconListIter icon;
	for(icon = _iconList.begin(); icon != _iconList.end(); icon++) {
		switch(_iconDirection) {
			case DD_RIGHT:
			case DD_LEFT:
				x = (i % _iconWrapCount);
				y = (i / _iconWrapCount);
				if(_iconWrapDirection == DD_UP) {
					y = -y;
				}
				if(_iconDirection == DD_LEFT) {
					x = -x;
				}
				break;

			case DD_DOWN:
			case DD_UP:
				x = (i / _iconWrapCount);
				y = (i % _iconWrapCount);
				if(_iconWrapDirection == DD_LEFT) {
					x = -x;
				}
				if(_iconDirection == DD_UP) {
					y = -y;
				}
				break;

		}

		if(!_hideText) {
			(*icon)->x = _iconX + ((boxWidth + _iconSpacingX) * x);
			(*icon)->y = _iconY + ((boxHeight + _iconSpacingY) * y);
			SetRect(&(*icon)->textRect, (*icon)->x + offsetTextX, (*icon)->y + offsetTextY, (*icon)->x + offsetTextX + _textWidth, (*icon)->y + offsetTextY + _textHeight);
			SetRect(&(*icon)->iconRect, (*icon)->x + offsetIconX, (*icon)->y + offsetIconY, (*icon)->x + offsetIconX + _iconSize, (*icon)->y + offsetIconY + _iconSize);
		}
		else {
			(*icon)->x = _iconX + ((_iconSize + _iconSpacingX) * x);
			(*icon)->y = _iconY + ((_iconSize + _iconSpacingY) * y);

			SetRect(&(*icon)->iconRect, (*icon)->x, (*icon)->y, (*icon)->x + _iconSize, (*icon)->y + _iconSize);
		}

		// Load icon
		(*icon)->SetIcon(_getIconFromPath((*icon)->filename, _iconSize, _iconSize));
		if(!(*icon)->GetIcon()) {
			LogPrintf(LOG_ERROR, "Could not load icon for \"%s\"", (*icon)->filename);
		}

		// Create buffer copy with fx changes entered

		if(_effectFlags & FX_ANY) {
			PAINTSTRUCT ps;
			HDC	hdcScreen = BeginPaint(_parentWnd, &ps);
			if((*icon)->GetIcon()) {
				(*icon)->hdcBuffer = CreateCompatibleDC(hdcScreen);
				(*icon)->hbmBuffer = CreateCompatibleBitmap(hdcScreen, _iconSize, _iconSize);
				SelectObject((*icon)->hdcBuffer, (*icon)->hbmBuffer);

				int y = 0;
				HDC hdc = CreateCompatibleDC((*icon)->hdcBuffer);
				HBITMAP hbmNew = CreateCompatibleBitmap((*icon)->hdcBuffer, 2 * _iconSize, _iconSize);
				SelectObject(hdc, hbmNew);

				// Temporarily draw up the icon to read it's pixels
				DrawIconEx(hdc, 0, 0, (*icon)->GetIcon(), _iconSize, _iconSize, 0, NULL, DI_NORMAL);
				// the mask of the icon
				DrawIconEx(hdc, _iconSize, 0, (*icon)->GetIcon(), _iconSize, _iconSize, 0, NULL, DI_MASK);

				while(y < _iconSize) {
					int x = 0;
					while(x < _iconSize) {
						// Loop through all transparent pixels...
						while(x < _iconSize) {
							// If the mask-pixel is white, then break
							if(!GetPixel(hdc, x + _iconSize, y)) {
								COLORREF cl = GetPixel(hdc, x, y);
								BYTE r = GetRValue(cl);
								BYTE g = GetGValue(cl);
								BYTE b = GetBValue(cl);

								// Saturation effect
								if(_effectFlags & FX_SATURATION) {
									BYTE gray = (BYTE)(r * 0.3086 + g * 0.6094 + b * 0.0820);

									r = (BYTE)((r * _saturation + gray * (255 - _saturation) + 255) >> 8);
									g = (BYTE)((g * _saturation + gray * (255 - _saturation) + 255) >> 8);
									b = (BYTE)((b * _saturation + gray * (255 - _saturation) + 255) >> 8);
								}
								// icon hue effect
								if(_effectFlags & FX_HUE) {
									// The B & R values of the hue is swapped here, can somebody tell me why it only works that way?
									// test normal
									r = (BYTE)((r * (255 - _hueIntensity) + GetRValue(_hueColor) * _hueIntensity + 255) >> 8);
									g = (BYTE)((g * (255 - _hueIntensity) + GetGValue(_hueColor) * _hueIntensity + 255) >> 8);
									b = (BYTE)((b * (255 - _hueIntensity) + GetBValue(_hueColor) * _hueIntensity + 255) >> 8);
								}
								SetPixel((*icon)->hdcBuffer, x, y, RGB(r, g, b));
							}
							// for transparent blt?
							else {
								SetPixel((*icon)->hdcBuffer, x, y, RGB(255, 0, 255));
							}
							x++;
						}
					}
					y++;
				}
				DeleteObject(hbmNew);
				DeleteDC(hdc);
			}
			DeleteObject(hdcScreen);
			EndPaint(_parentWnd, &ps);
		}

		i++;
	}

	_loadTooltips();
}

bool Group::_compareIconsPrio(Icon* i1, Icon* i2)
{
	return (((i1->lFlags & Icon::ICON_FIRST) && !(i2->lFlags & Icon::ICON_FIRST)) || (!(i1->lFlags & Icon::ICON_LAST) && (i2->lFlags & Icon::ICON_LAST)));
	//return ((i1->lFlags & Icon::ICON_FIRST) > (i2->lFlags & Icon::ICON_FIRST) || (i1->lFlags & Icon::ICON_LAST) > (i2->lFlags & Icon::ICON_LAST));
};

bool Group::_compareIconsFolder(Icon* i1, Icon* i2)
{
	return ((i1->lFlags & Icon::ICON_FOLDER) > (i2->lFlags & Icon::ICON_FOLDER));
};

bool Group::_compareIcons(Icon* i1, Icon* i2)
{
	return (stricmp(i1->title, i2->title) < 0);
};

void Group::paintIcons(HDC hDC)
{
	if(_visible) {
		if(_iconsLoaded == FALSE) {
			_loadIcons();
		}

		COLORREF oldColor = SetTextColor(hDC, _textColor);
		HFONT oldFont = (HFONT)SelectObject(hDC, _textFont);
		int oldMode = SetBkMode(hDC, TRANSPARENT);

		for(IconListIter it = _iconList.begin(); it != _iconList.end(); it++) {
			//DrawFocusRect(hDC, &(*it)->textRect);
			//DrawFocusRect(hDC, &(*it)->iconRect);
			if((*it)->GetIcon()) {
				if(_effectFlags & FX_ANY) {
					// Paint from the buffer
					TransparentBltLS(hDC, (*it)->iconRect.left, (*it)->iconRect.top, _iconSize, _iconSize, (*it)->hdcBuffer, 0, 0, RGB(255, 0, 255));
				}
				else {
					DrawIconEx(hDC, (*it)->iconRect.left, (*it)->iconRect.top, (*it)->GetIcon(), _iconSize, _iconSize, 0, NULL, DI_NORMAL);
				}
			}
			if(!_hideText) {
				if(_textBackground) {
					FillRect(hDC, &(*it)->textRect, _textBackground);
				}
				if(_selectedIcon == (*it)) {
					DrawFocusRect(hDC, &(*it)->textRect);
				}

				if(_textShadowDepth) {
					RECT r;
					CopyRect(&r, &(*it)->textRect);
					OffsetRect(&r, _textShadowDepth, _textShadowDepth);
					SetTextColor(hDC, _textShadowColor);
					DrawText(hDC, (*it)->title, strlen((*it)->title), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);

					if(_textOutline) {
						OffsetRect(&r, -1, 0);
						DrawText(hDC, (*it)->title, strlen((*it)->title), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
						OffsetRect(&r, -1, 0);
						DrawText(hDC, (*it)->title, strlen((*it)->title), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
						OffsetRect(&r, 0, -1);
						DrawText(hDC, (*it)->title, strlen((*it)->title), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
						OffsetRect(&r, 0, -1);
						DrawText(hDC, (*it)->title, strlen((*it)->title), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
						OffsetRect(&r, 1, 0);
						DrawText(hDC, (*it)->title, strlen((*it)->title), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
						OffsetRect(&r, 1, 0);
						DrawText(hDC, (*it)->title, strlen((*it)->title), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
						OffsetRect(&r, 0, 1);
						DrawText(hDC, (*it)->title, strlen((*it)->title), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
						SetTextColor(hDC, _textColor);
					}
					SetTextColor(hDC, _textColor);
				}

				if(_textOutline) {
					RECT r;
					CopyRect(&r, &(*it)->textRect);
					OffsetRect(&r, 1, 1);
					SetTextColor(hDC, _textOutlineColor);
					DrawText(hDC, (*it)->title, strlen((*it)->title), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
					OffsetRect(&r, -1, 0);
					DrawText(hDC, (*it)->title, strlen((*it)->title), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
					OffsetRect(&r, -1, 0);
					DrawText(hDC, (*it)->title, strlen((*it)->title), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
					OffsetRect(&r, 0, -1);
					DrawText(hDC, (*it)->title, strlen((*it)->title), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
					OffsetRect(&r, 0, -1);
					DrawText(hDC, (*it)->title, strlen((*it)->title), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
					OffsetRect(&r, 1, 0);
					DrawText(hDC, (*it)->title, strlen((*it)->title), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
					OffsetRect(&r, 1, 0);
					DrawText(hDC, (*it)->title, strlen((*it)->title), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
					OffsetRect(&r, 0, 1);
					DrawText(hDC, (*it)->title, strlen((*it)->title), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
					SetTextColor(hDC, _textColor);
				}
				DrawText(hDC, (*it)->title, strlen((*it)->title), &(*it)->textRect, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
			}
		}

		SetTextColor(hDC, oldColor);
		SelectObject(hDC, oldFont);
		SetBkMode(hDC, oldMode);
	}
}

Icon* Group::_getIconAtPt(int x, int y)
{
	for(IconListIter it = _iconList.begin(); it != _iconList.end(); it++) {
		if((*it)->PointOver(x, y)) {
			return (*it);
		}
	}
	return NULL;
}

HICON Group::_getIconFromPath(LPCSTR pszPath, int cxDesired, int cyDesired)
{
	if(!PrivateExtractIcons) {
		return NULL;
	}

	HICON hIcon = NULL;
	LPCSTR pszExt;
    pszExt = strrchr(pszPath, '.');

    if(pszExt && (!stricmp(pszExt, ".ico") || !stricmp(pszExt, ".cur") || !stricmp(pszExt, ".ani") || !stricmp(pszExt, ".exe"))) {
		// Executables, icons and cursors have their own icons
		PrivateExtractIcons(pszPath, 0, cxDesired, cyDesired, &hIcon, NULL, 1, NULL);
    }
    else if(pszExt && !stricmp(pszExt, ".lnk")) {
		// Special case for link files
		// If the icon is set for the link, retrieve it
		SHFILEINFO shfi;
		SHGetFileInfo(pszPath, 0, &shfi, sizeof(SHFILEINFO), SHGFI_ICONLOCATION);
		// Extract the icon
		if(PrivateExtractIcons(shfi.szDisplayName, shfi.iIcon, cxDesired, cyDesired, &hIcon, NULL, 1, NULL) == 0xFFFFFFFF) {
			// If the icon was not explicitly set in the linkfile, we'll use the icon from the target file
			HRESULT hRes;				
			char szInfo[1024];
			IShellLink* psl;

			// Create the ShellLink object
			hRes = CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER, IID_IShellLink, (LPVOID*)&psl);
			if(SUCCEEDED(hRes)) {
				IPersistFile* ppf;
				// Bind the ShellLink object to the Persistent File
				hRes = psl->QueryInterface(IID_IPersistFile, (LPVOID *) &ppf);
				if(SUCCEEDED(hRes)) {
					WORD wsz[MAX_PATH];

					MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, pszPath, -1, wsz, MAX_PATH);
					hRes = ppf->Load(wsz, 0);
					
					if(SUCCEEDED(hRes)) {
						// Retrieve the target of the link file
						psl->GetPath(szInfo, 1024, NULL, SLGP_UNCPRIORITY);
					}
				}
			}
			psl->Release();

			// Extract the icon from the link target
			//PrivateExtractIcons(szInfo, 0, cxDesired, cyDesired, &hIcon, NULL, 1, NULL);
			hIcon = _getIconFromPath(szInfo, cxDesired, cyDesired);
		}
    }
    else {
		// Load icons for all other file types
		if(pszExt) {
			// Retrieve the icon associated with the file type from the registry
			char szFileType[1024], szIcon[1024];
			int iIconIndex = 0;
			DWORD dwDataSize = 1024;
			// Get the file type from the extension
			if(SHGetValue(HKEY_CLASSES_ROOT, pszExt, NULL, NULL, szFileType, &dwDataSize) == ERROR_SUCCESS) {
				strcat(szFileType, "\\DefaultIcon");
				dwDataSize = 1024;
				// Get the icon
				if(SHGetValue(HKEY_CLASSES_ROOT, szFileType, NULL, NULL, szIcon, &dwDataSize) == ERROR_SUCCESS) {
					char* pszSplit = strrchr(szIcon, ',');
					if(pszSplit) {
						iIconIndex = atoi(pszSplit + 1);
						(*pszSplit) = '\0';
					}

					// Extract the icon
					PrivateExtractIcons(szIcon, iIconIndex, cxDesired, cyDesired, &hIcon, NULL, 1, NULL);
				}
			}

			if(!hIcon) {
				// No icon
				PrivateExtractIcons("shell32.dll", 0, cxDesired, cyDesired, &hIcon, NULL, 1, NULL);
			}
		}
		else {
			// Let the shell decide the icon
			SHFILEINFO shfi;
			SHGetFileInfo(pszPath, 0, &shfi, sizeof(SHFILEINFO), SHGFI_ICONLOCATION);
			// Extract the icon
			PrivateExtractIcons(shfi.szDisplayName, shfi.iIcon, cxDesired, cyDesired, &hIcon, NULL, 1, NULL);
		}
    }

	// If all else fails
	if(hIcon == NULL) {
		SHFILEINFO shfi;
		SHGetFileInfo(pszPath, 0, &shfi, sizeof(SHFILEINFO), SHGFI_ICON | (cxDesired <= 24 ? SHGFI_SMALLICON : 0));
		return shfi.hIcon;
	}

	return hIcon;
}

void Group::_StrRetToBuf(LPSTRRET pstr, LPCITEMIDLIST pidl, LPTSTR pszBuf, UINT cchBuf)
{
	switch(pstr->uType) {
		case STRRET_WSTR:
			{
				WideCharToMultiByte(CP_ACP, 0, pstr->pOleStr, -1, pszBuf, cchBuf, NULL, NULL);
				LPMALLOC pMalloc;
				if(SUCCEEDED(SHGetMalloc(&pMalloc))) {
					pMalloc->Free(pstr->pOleStr);
					pMalloc->Release();
				}
			}
			break;

		case STRRET_CSTR:
			lstrcpyn(pszBuf, pstr->cStr, cchBuf);
			break;

		case STRRET_OFFSET:
			lstrcpyn(pszBuf, (LPTSTR)pidl + pstr->uOffset, cchBuf);
			break;

	}
}

void Group::_getString(const char* key, char* value, const char* defaultValue, int maxLen)
{
	char buffer[MAX_RCCOMMAND] = "";
	sprintf(buffer, "IconDesk%s%s", name, key);
	GetRCString(buffer, value, defaultValue, maxLen);
}

BOOL Group::_getBool(const char* key, BOOL defaultValue)
{
	char buffer[MAX_RCCOMMAND] = "";
	sprintf(buffer, "IconDesk%s%s", name, key);
	return GetRCBoolDef(buffer, defaultValue);
}

int Group::_getInt(const char* key, int defaultValue)
{
	char buffer[MAX_RCCOMMAND] = "";
	sprintf(buffer, "IconDesk%s%s", name, key);
	return GetRCInt(buffer, defaultValue);
}

COLORREF Group::_getColor(const char* key, int defaultValue)
{
	char buffer[MAX_RCCOMMAND] = "";
	sprintf(buffer, "IconDesk%s%s", name, key);
	return GetRCColor(buffer, defaultValue);
}
